class Node:
    def __init__(self,data) -> None:
        self.data = data
        self.prev = None
        self.next = None

class DLL:
    def __init__(self) -> None:
        self.head = None
        self.tail = None
    
    def insert_first(self,data):
        new_node = Node(data)
        if self.head is None and self.tail is None:
            self.head = self.tail = new_node
        else:
            new_node.next = self.head
            self.head.prev = new_node
            self.head = new_node
    
    def insert_last(self,data):
        new_node = Node(data)
        if self.head is None and self.tail is None:
            self.head = self.tail = new_node
        else:
            self.tail.next = new_node
            new_node.prev = self.tail
            self.tail = new_node
    
    def insert_at_pos(self,pos,data):
        new_node = Node(data)
        if self.head is None and self.tail is None:
            self.head = self.tail = new_node
        elif pos < 0:
            raise Exception("Position should be greater than zero")
        elif pos == 0:
            new_node.next = self.head
            self.head.prev = new_node
            self.head = new_node
        else: # pos > 0
            curr = self.head
            while pos-1 != 0 and curr.next is not None:
                curr = curr.next
                pos -= 1
            if curr.next is not None:
                dummy = curr.next
                curr.next = new_node
                new_node.prev = curr
                new_node.next = dummy
                dummy.prev = new_node
            else: # curr.next is None
                curr.next = new_node
                new_node.prev = curr
                self.tail = new_node
                
    def traverse_forward(self):
        curr = self.head
        output = []
        while curr is not None:
            output.append(curr.data)
            curr = curr.next
        print(output)
        
    def traverse_backward(self):
        curr = self.tail
        output = []
        while curr is not None:
            output.append(curr.data)
            curr = curr.prev
        print(output)
    
    def search(self,element):
        index = 0
        curr = self.head
        while curr is not None:
            if curr.data == element:
                return index
            else:
                curr = curr.next
                index += 1
        return -1
            
    def delete_first(self):
        if self.head is None and self.tail is None:
            raise Exception("Cannot delete from empty doubly linked list")
        else:
            to_delete = self.head
            if self.head.next is None and self.tail.next is None:
                self.head = self.tail = None 
            else: # self.head.next is not None
                self.head = self.head.next 
            del to_delete
                
    def delete_last(self):
        if self.head is None and self.tail is None:
            raise Exception("Cannot delete from empty doubly linked list")
        else:
            if self.head.next is None and self.tail.next is None:
                self.head = self.tail = None 
            else:
                previous = None
                curr = self.head
                while curr.next is not None:
                    previous = curr
                    curr = curr.next
                if previous is not None:
                    previous.next = None
                    self.tail = previous
                else:
                    self.head = self.tail = None
                del curr 
    
    def delete(self,element):
        if self.head is None and self.tail is None:
            raise Exception("Cannot delete from empty doubly linked list")
        if self.head.data == element:
            to_delete = self.head
            self.head = self.head.next
            if self.head is not None:
               self.head.prev = None
            else:
                self.tail = None
            del to_delete
        else:
            previous = None
            curr = self.head
            while curr is not None:
                if curr.data != element:
                    previous = curr
                    curr = curr.next
                else:
                    break
            if curr is not None:
                to_delete = curr 
                if curr.next is not None:
                    dummy = curr.next
                    previous.next = dummy
                    dummy.prev = previous
                else:
                    previous.next = None
                    self.tail = previous
                del to_delete
            else:
                 raise Exception("The element is not present in the doubly linked list")
    
    def delete_at_pos(self,pos):
        if self.head is None:
            raise Exception("Cannot delete from empty doubly linked list")
        if pos < 0:
            raise Exception("The position should be greater than zero")
        if pos == 0:
            to_delete = self.head
            self.head = self.head.next
            if self.head is not None:
               self.head.prev = None
            else:
                self.tail = None
            del to_delete
        if pos > 0:
            curr = self.head
            while pos-1 != 0 and curr.next is not None:
                curr = curr.next
                pos -= 1
            if curr.next is not None:
                to_delete = curr.next 
                if curr.next.next is not None:
                    dummy = curr.next.next
                    curr.next = dummy
                    dummy.prev = curr
                else:
                    curr.next = None 
                    self.tail = curr 
                del to_delete
            else:
                raise Exception("The position should be less than size of linkedlist")
    
    def reverse(self):
        dummy = None
        curr = self.head
        future_tail = self.head
        while curr is not None:
            dummy = curr.prev
            curr.prev = curr.next
            curr.next = dummy
            curr = curr.prev
        if dummy is not None:
            self.head = dummy.prev
            self.tail = future_tail
    
    def sort(self):
        def getMiddle(dummy):
            if (dummy == None):
                return dummy
            slow = dummy
            fast = dummy
            while (fast.next != None and fast.next.next != None):
                slow = slow.next
                fast = fast.next.next
            tup = dummy, slow.next
            slow.next = None
            return tup
        
        def sortedMerge(a,b):
            if a is None:
                return b
            if b is None:
                return a
            if a.data < b.data:
                a.next = sortedMerge(a.next,b)
                a.next.prev = a
                a.prev = None
                return a
            else:
                b.next = sortedMerge(a,b.next)
                b.next.prev = a
                b.prev = None
                return b
        
        def find_tail():
            curr = self.head
            while curr.next is not None:
                curr = curr.next
            return curr
            
        def mergesort(dummy):
            if dummy is None or dummy.next is None:
                return dummy
            front, back = getMiddle(dummy)
            front = mergesort(front)
            back = mergesort(back)
            return sortedMerge(front, back)
        
        sorted_head = mergesort(self.head)
        self.head = sorted_head
        self.tail = find_tail()
        
    def size(self):
        curr = self.head
        count = 0
        while curr is not None:
            count += 1
            curr = curr.next 
        return count
               
    def __str__(self):
        return "<class 'DLL'>"
    
    def first_element(self):
        if self.head is not None:
            return self.head.data
        else:
            raise Exception("No items in the doubly linked list")
    
    def last_element(self):
        if self.head is not None and self.tail is not None:
            return self.tail.data
        else:
            raise Exception("No items in the doubly linked list")


















class Node:
    def __init__(self,data) -> None:
        self.data = data 
        self.next = None 
        
class SCLL:
    def __init__(self) -> None:
        self.head = None 
    
    def insert_first(self,data):
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
            new_node.next = new_node
        else:
            new_node.next = self.head 
            curr = self.head
            while curr.next is not self.head:
                curr = curr.next 
            curr.next = new_node
            self.head = new_node
            
    
    def insert_last(self,data):
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
            new_node.next = new_node
        else:
            curr = self.head
            while curr.next is not self.head:
                curr = curr.next
            curr.next = new_node
            new_node.next = self.head
    
    def insert_at_pos(self,pos, data):
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
            new_node.next = new_node
        elif pos < 0:
            raise Exception("Position should be greater than zero")
        elif pos == 0:
            new_node.next = self.head
            curr = self.head
            while curr.next is not self.head:
                curr = curr.next
            curr.next = new_node
            self.head = new_node
        else: # pos > 0
            curr = self.head
            while pos-1 != 0 and curr.next is not self.head:
                curr = curr.next
                pos -= 1
            if curr.next is not self.head:
                dummy = curr.next
                curr.next = new_node
                new_node.next = dummy
            else: # curr.next is self.head
                curr.next = new_node
                new_node.next = self.head
                
    def traverse(self):
        curr = self.head
        output = []
        while curr is not None:
            output.append(curr.data)
            if curr.next is not self.head:
               curr = curr.next
            else:
                break
        print(output)
        
    def search(self,element):
        curr = self.head
        index = 0
        while True:
            if curr.data == element:
                return index
            curr = curr.next 
            index += 1   
            if curr == self.head:
                break 
        return -1       
    
    def delete_first(self):
        if self.head is None:
            raise Exception("Cannot delete from empty Single circular linked list")
        else:
            if self.head.next is not self.head:
                curr = self.head
                while curr.next is not self.head:
                   curr = curr.next 
                curr.next = self.head.next 
                self.head = self.head.next 
            else:
                self.head = None 
            
    def delete_last(self):
        if self.head is None:
            raise Exception("Cannot delet from empty Single circular linked list")
        else:
            if self.head.next is not self.head:
                previous = None 
                curr = self.head
                while curr.next is not self.head:
                   previous = curr
                   curr = curr.next 
                previous.next = self.head 
            else:
                self.head = None 
    
    def delete(self,element):
        if self.head is None:
            raise Exception("Cannot delete from empty Single circular linked list")
        else:
            previous = None 
            curr = self.head
            while curr.next is not self.head and curr.data != element:
                previous = curr
                curr = curr.next 
            if previous is not None:
                previous.next = curr.next 
            else:
                temp = self.head
                while temp.next is not self.head:
                    temp = temp.next
                if self.head.next is not self.head: 
                    temp.next = self.head.next 
                    self.head = self.head.next
                else:
                    if self.head.data == element:
                       self.head = None 
                    else:
                        raise Exception("The element is not present in the Single circular linked list")

                
    def delete_at_pos(self,pos):
        if self.head is None:
            raise Exception("Cannot delete from empty Single circular linked list")
        if pos < 0:
            raise Exception("The position should be greater than zero")
        if pos == 0:
            if self.head.next is not self.head:
                curr = self.head
                while curr.next is not self.head:
                    curr = curr.next 
                curr.next = self.head.next 
                self.head = self.head.next 
            else:
                self.head = None 
        if pos > 0:
            curr = self.head
            while pos-1 != 0 and curr.next is not self.head:
                curr = curr.next 
                pos -= 1
            if curr.next is not self.head:
                dummy = curr.next.next 
                if dummy is not self.head:
                    curr.next = dummy
                else:
                    curr.next = self.head
            else:
                raise Exception("The position should be less than size of linkedlist") 
                
    def reverse(self):
        if self.head is not None:
            prev = None
            curr = self.head
            after = curr.next 
            curr.next = prev 
            prev = curr
            curr = after
            while curr is not self.head:
                after = curr.next 
                curr.next = prev 
                prev = curr
                curr = after 
            self.head.next = prev 
            self.head = prev 
    
    def sort(self):
        def remove_connection(dummy):
            curr = dummy
            while curr.next is not self.head:
                curr = curr.next 
            curr.next = None 
            return dummy 
        
        def make_connection(dummy):
            curr = dummy
            while curr.next is not None:
                curr = curr.next 
            curr.next = self.head 
        
        def get_middle(dummy):
            if (dummy == None):
                return dummy
            slow = dummy
            fast = dummy
            while (fast.next != None and fast.next.next != None):
                slow = slow.next
                fast = fast.next.next
            tup = dummy, slow.next
            slow.next = None
            return tup
        
        def sorted_merge(a,b):
            if a is None:
                return b
            if b is None:
                return a 
            if a.data <= b.data:
                result = a
                result.next = sorted_merge(a.next,b)
            else:
                result = b
                result.next = sorted_merge(a,b.next)
            return result
        
        def merge_sort(dummy):
            if dummy is None or dummy.next is None:
                return dummy
            else:
                front, back = get_middle(dummy)
                front = merge_sort(front)
                back = merge_sort(back)
                return sorted_merge(front,back)
        
        sll_head = remove_connection(self.head)
        final_head = merge_sort(sll_head)
        self.head = final_head
        make_connection(self.head)
        
    def size(self):
        count = 0
        curr = self.head
        while curr is not None:
            count += 1
            if curr.next is not self.head:
                curr = curr.next
            else:
                break 
        return count
    
    def __str__(self):
        return "<class 'SCLL'>"
    
        
scll = SCLL()
scll.insert_first("boy")
scll.insert_last("auto")
scll.traverse()
scll.sort()
scll.traverse()








from PyMolecules import SingleLinkedList, DoubleLinkedList,SingleCircularLinkedList


__author__ = "Yokesh Bollineni"
__version__ = "2.0.1"
__what_is_new__ = "Added SCLL class, Bug fix in size() method of DLL class"



from typing import Optional


class Node:
    def __init__(self, data) -> None:
        self.data = data
        self.next = None


class SLL:
    def __init__(self) -> None:
        self.head = None

    def insert_first(self, data) -> None:
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
        else:
            new_node.next = self.head
            self.head = new_node

    def insert_last(self, data) -> None:
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
        else:
            curr = self.head
            while curr.next is not None:
                curr = curr.next
            curr.next = new_node

    def insert_at_pos(self, pos, data) -> None:  # index starts from '0'
        new_node = Node(data)
        if self.head is None:
            self.head = new_node
            return 
        if pos < 0:
            raise Exception("Position should be greater than zero")
        if pos == 0:
            new_node.next = self.head
            self.head = new_node
        elif pos > 0:
            curr = self.head
            while pos - 1 != 0 and curr.next is not None:
                curr = curr.next
                pos -= 1
            after = curr.next
            curr.next = new_node
            new_node.next = after

    def traverse(self) -> None:
        curr = self.head
        output = []
        while curr is not None:
            output.append(curr.data)
            curr = curr.next
        print(output)

    def search(self, element) -> Optional[int]:
        curr = self.head
        index = 0
        while curr is not None:
            if curr.data == element:
                return index
            else:
                curr = curr.next
                index += 1
        return -1  # If element is not found

    def delete_first(self) -> None:
        if self.head is not None:
            self.head = self.head.next
        else:
            raise Exception("Cannot delete from empty linkedlist")

    def delete_last(self) -> None:
        if self.head is not None:
            curr = self.head
            prev = None
            while curr.next is not None:
                prev = curr
                curr = curr.next
            if prev is not None:
                prev.next = None
            else:
                self.head = None
        else:
            raise Exception("Cannot delete from empty linkedlist")

    def delete(self, element) -> None:
        if self.head is None:
            raise Exception("Cannot delete from empty linkedlist")
        if self.head.data == element:
            to_delete = self.head
            self.head = self.head.next
            del to_delete
        else:
            prev = None
            curr = self.head
            while curr is not None:
                if curr.data != element:
                    prev = curr
                    curr = curr.next
                else:
                    break
            if curr is not None:
                to_del = curr
                prev.next = curr.next
                del to_del
            else:
                raise Exception("The element is not present in the linkedlist")

    def delete_at_pos(self, pos) -> None:
        if self.head is None:
            raise Exception("Cannot delete from empty linkedlist")
        if pos < 0:
            raise Exception("The position should be greater than zero")
        if pos == 0:
            to_delete = self.head
            self.head = self.head.next
            del to_delete
        if pos > 0:
            curr = self.head
            while pos - 1 != 0 and curr.next is not None:
                curr = curr.next
                pos -= 1
            if curr.next is not None:
                to_delete = curr.next 
                curr.next = curr.next.next
                del to_delete
            else:
                raise Exception("The position should be less than size of linkedlist")

    def reverse(self) -> None:
        prev = None
        curr = self.head
        while curr is not None:
            after = curr.next
            curr.next = prev
            prev = curr
            curr = after
        self.head = prev

    def sort(self) -> None:
        def getMiddle(dummy):
            if (dummy == None):
                return dummy
            slow = dummy
            fast = dummy
            while (fast.next != None and fast.next.next != None):
                slow = slow.next
                fast = fast.next.next
            tup = dummy, slow.next
            slow.next = None
            return tup

        def sortedMerge(a, b):
            if a is None:
                return b
            elif b is None:
                return a
            if a.data <= b.data:
                result = a
                result.next = sortedMerge(a.next, b)
            else:
                result = b
                result.next = sortedMerge(a, b.next)
            return result

        def mergesort(dummy):
            if dummy is None or dummy.next is None:
                return dummy
            front, back = getMiddle(dummy)
            front = mergesort(front)
            back = mergesort(back)
            return sortedMerge(front, back)

        sorted_head = mergesort(self.head)
        self.head = sorted_head

    def size(self) -> int:
        curr = self.head
        count = 0
        while curr is not None:
            count += 1
            curr = curr.next
        return count

    def __str__(self):
        return "<class 'SLL'>"





from setuptools import setup, find_packages

setup(
    name='PyMolecules',
    version='2.0.1',
    description='A Python package for all types of linked lists and most useful methods',
    long_description=open('README.txt').read()+'\n\n'+open('CHANGELOG.txt').read(),
    url='',
    author='Yokesh Bollineni',
    author_email='bollineniyokesh@gmail.com',
    license='MIT',
    keywords='linked lists',
    packages=find_packages(),
    install_requires=['']
)

